/*
	Vertris v0.3.2
	Copyright (C) 2006 - 2007, Andrs Rassol
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <string.h>

#include <allegro.h>

#include "vertris.h"

SAMPLE *Lineclear, *Move;
extern Vertromino Vertro;

extern Vertromino Vertro_Next;

extern char Vertris[VERTRIS_W][VERTRIS_H], Next_Vertro, Fullscreen, Pause, Sound;
char FXVertris[VERTRIS_W][VERTRIS_H];

extern char Show_HiScores, Show_Name_Input, Name[NAME_LENGTH];

extern unsigned int Lines, Max_Lines[MAX_HISCORES], Score, Max_Score[MAX_HISCORES], Level;
extern char Max_Name[MAX_HISCORES][NAME_LENGTH];
extern char Name[NAME_LENGTH];
static int Last_Score, Last_Vertro;

static BITMAP *Buf, *bg, *vertrominos[VERTROMINOS];
static FONT *VFont;

volatile int FX;

int Redraw_All;

extern char Ghosting_FX;

static void add_fx() {FX++;} END_OF_STATIC_FUNCTION(add_tick);

static void reset_fx() {FX = 0;} END_OF_FUNCTION(reset_ticks);

static int depth_select(int depth)
{
	set_color_depth(depth);

	if (Fullscreen)
		return set_gfx_mode(GFX_AUTODETECT_FULLSCREEN, WIDTH, HEIGHT, 0, 0) == 0;
	else
		return set_gfx_mode(GFX_AUTODETECT_WINDOWED, WIDTH, HEIGHT, 0, 0) == 0;
}

void shutdown_output()
{
	unsigned int i;

	destroy_bitmap(Buf);
	destroy_bitmap(bg);
	
	for (i = VERTROMINOS; i--; )
		destroy_bitmap(vertrominos[i]);

	destroy_font(VFont);

	destroy_sample(Lineclear);
	destroy_sample(Move);
}

static void redraw_all() {Redraw_All = 1;}

static void do_pause() {Pause = 1;}

void init_output()
{	
	unsigned int i;
	int error;

	error = 0;
	
	if (!(depth_select(32) || depth_select(24) || depth_select(16) || depth_select(15) || depth_select(8)))
	{
		allegro_message("Error setting graphics mode\n%s\n", allegro_error);
		exit(1);
	}	

	Lineclear = load_sample("sounds/lineclear.wav");
	Move = load_sample("sounds/move.wav");
	
	if (!Lineclear || !Move)
	{
		Sound = 0;
		destroy_sample(Lineclear);
		destroy_sample(Move);
	}
	
	Buf = create_bitmap(SCREEN_W, SCREEN_H);
	bg = load_bmp("images/Background.bmp", NULL);

	if (!bg)
		error = 1;

	vertrominos[0] = load_bmp("images/I.bmp", NULL);
	vertrominos[1] = load_bmp("images/T.bmp", NULL);
	vertrominos[2] = load_bmp("images/O.bmp", NULL);
	vertrominos[3] = load_bmp("images/L.bmp", NULL);
	vertrominos[4] = load_bmp("images/J.bmp", NULL);
	vertrominos[5] = load_bmp("images/S.bmp", NULL);
	vertrominos[6] = load_bmp("images/Z.bmp", NULL);

	for (i = VERTROMINOS; i--; )
		if (!vertrominos[i])
			error = 1;

	if (error)
	{
		allegro_message("Error loading graphics\n");
		shutdown_output();
		exit(1);
	}

	VFont = load_font("images/font.pcx", NULL, NULL);
	if (!VFont)
	{
		VFont = font;
		allegro_message("Error loading font\n");
	}

	textout_centre_ex (bg, VFont, "Vertris v0.3.2", SCREEN_W/2, 10, -1, -1);

	textout_right_ex(bg, VFont, "Copyright  2006 - 2007, Andrs Rassol", SCREEN_W - 5, SCREEN_H - 25, -1, -1);

	textout_ex (bg, VFont, "Keys", 430, 35, -1, -1);
	textout_ex (bg, VFont, "Left, Right, Down: Move", 430, 55, -1, -1);
	textout_ex (bg, VFont, "Up: Rotate Left", 430, 75, -1, -1);
	textout_ex (bg, VFont, "Numpad 0: Rotate Right", 430, 95, -1, -1);
	textout_ex (bg, VFont, "Space: Instant Drop", 430, 115, -1, -1);

	textout_ex (bg, VFont, "Alt + Enter: Fullscreen", 430, 155, -1, -1);
	textout_ex (bg, VFont, "R: Restart", 430, 175, -1, -1);
	textout_ex (bg, VFont, "P: Pause", 430, 195, -1, -1);
	textout_ex (bg, VFont, "Esc: Exit", 430, 215, -1, -1);

	Redraw_All = 1;

	set_display_switch_callback(SWITCH_IN, redraw_all);
	set_display_switch_callback(SWITCH_OUT, do_pause);

	set_trans_blender(255, 255, 255, 128);

	LOCK_FUNCTION(add_fx);
	LOCK_FUNCTION(reset_fx);
	LOCK_VARIABLE(FX);
	install_int(add_fx, 250);
}

static void textouts()
{
	int i;
	int x, y;

	if (Pause)
		textout_centre_ex (Buf, VFont, "Pause", SCREEN_W/2, SCREEN_H/2 - 5, -1, -1);

	textprintf_ex(Buf, VFont, 10, 35, -1, -1, "Lines: %d", Lines);

	textprintf_ex(Buf, VFont, 10, 55, -1, -1, "Score: %d", Score);

	textout_ex(Buf, VFont, "Next:", 10, 75, -1, -1);

	if (Show_Name_Input)
	{
		x = SCREEN_W / 2 - VERTRIS_W * SQUARE_SIZE / 2 - 20;

		textout_ex(Buf, VFont, "Please enter your name:", x, 60, -1, -1);

		textprintf_ex(Buf, VFont, x, 80, -1, -1, "%s", Name);
	}
	else if (Show_HiScores)
	{
		x = 10;
		y = 200;

		textout_ex(Buf, VFont, "Name", x, y, -1, -1);
		textout_ex(Buf, VFont, "Score", x + 100, y, -1, -1);
		textout_ex(Buf, VFont, "Lines", x + 190, y, -1, -1);

		for (i = 0; i < MAX_HISCORES; i++)
		{
			if (Score == Max_Score[i] && Lines == Max_Lines[i] &&
			!ustrcmp(Name, Max_Name[i]))
			{
			 	textprintf_ex(Buf, VFont, x, y + (i+1) * 20, 0x00FF00, -1, "%s", Max_Name[i]);
				textprintf_ex(Buf, VFont, x + 100, y + (i+1) * 20, 0x00FF00, -1, "%d", Max_Score[i]);
				textprintf_ex(Buf, VFont, x + 190, y + (i+1) * 20, 0x00FF00, -1, "%d", Max_Lines[i]);
			}
			else
			{
				textprintf_ex(Buf, VFont, x, y + (i+1) * 20, -1, -1, "%s", Max_Name[i]);
				textprintf_ex(Buf, VFont, x + 100, y + (i+1) * 20, -1, -1, "%d", Max_Score[i]);
				textprintf_ex(Buf, VFont, x + 190, y + (i+1) * 20, -1, -1, "%d", Max_Lines[i]);
			}
		}

		textout_centre_ex(Buf, VFont, "Press <Enter> to", SCREEN_W/2, SCREEN_H/2 - 30, -1, -1);
		textout_centre_ex(Buf, VFont, "continue...", SCREEN_W/2, SCREEN_H/2 - 10, -1, -1);
	}
}

static void draw_vertro(char vertro_type, int x, int y, int fx)
{
	switch (vertro_type)
	{
		case 'I': vertro_type = 0; break;
		case 'T': vertro_type = 1; break;
		case 'O': vertro_type = 2; break;
		case 'L': vertro_type = 3; break;
		case 'J': vertro_type = 4; break;
		case 'S': vertro_type = 5; break;
		case 'Z': vertro_type = 6;
	}

	if (!fx)
		blit(vertrominos[vertro_type], Buf, 0, 0, x, y, SQUARE_SIZE, SQUARE_SIZE);
	else if (fx == 1)
		draw_gouraud_sprite(Buf, vertrominos[vertro_type], x, y, 8, 255, 255, 255);
	else if (fx == 2)
		draw_gouraud_sprite(Buf, vertrominos[vertro_type], x, y, 255, 8, 255, 8);
	else if (fx == 3)
		draw_gouraud_sprite(Buf, vertrominos[vertro_type], x, y, 255, 255, 8, 255);
	else if (fx == 4)
		draw_trans_sprite(Buf, vertrominos[vertro_type], x, y);
}

static void draw_next_vertro()
{
	unsigned int i, j;
	
	for (i = MAX_VERTROMINO - 1; i--; )
		for (j = MAX_VERTROMINO; j--; )
			if (Vertro_Next.shape[i][j])
				draw_vertro(Vertro_Next.shape[i][j], 20 + i * SQUARE_SIZE, 160 - j * SQUARE_SIZE, 0);
}

void play_sound(SAMPLE *sound)
{
	if (Sound)
		play_sample(sound, 255, 128, 1000, 0);	
}

inline void output_blit(BITMAP *source, BITMAP *dest)
{
	if (Redraw_All)
		blit(source, dest, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
	else
	{
		if (Score > Last_Score || Last_Vertro != Vertro_Next.type)
		{
			blit(source, dest, 60, 40, 60, 40, 140, 40);
			blit(source, dest, 20, 100, 20, 100, 60, 80);
		}

		blit(source, dest, BOARD_X, BOARD_Y - 400, BOARD_X, BOARD_Y - 400, 200, 401);
	}
}

inline void do_fx(int x, int y)
{
	if (AL_RAND() % 250 == 1 && !FXVertris[x][y])
		FXVertris[x][y] = 1;
	else if (FXVertris[x][y] >= 4)
		FXVertris[x][y] = 0;
	else if (FXVertris[x][y])
		FXVertris[x][y] = (FXVertris[x][y] + 1) % 3;
}

void output()
{
	unsigned int i, j, k;
	unsigned int x1, y1, x2, y2;

	output_blit(bg, Buf);

	if (!Pause)
	{
		for (i = MAX_VERTROMINO; i--; )
			for (j = MAX_VERTROMINO; j--; )
			{
				x2 = Vertro.x + i;
				y2 = Vertro.y + j;

				if (Vertro.shape[i][j] && x2 < VERTRIS_W && y2 < VERTRIS_H)
				{
					x1 = BOARD_X + SQUARE_SIZE * x2;
					y1 = BOARD_Y - SQUARE_SIZE * y2;

					for (k = Ghosting_FX; k--; )
					{
						set_trans_blender(255, 255, 255, 160 - 25 *  k);
						draw_vertro(Vertro.shape[i][j], x1, y1 - (k + 2) * SQUARE_SIZE, 4);
					}
					set_trans_blender(255, 255, 255, 128);

					draw_vertro(Vertro.shape[i][j], x1, y1 - SQUARE_SIZE, 0);

					hline(Buf, x1, BOARD_Y, x1 + SQUARE_SIZE - 1, 0xFFFFFF);
				}
			}

		for (i = VERTRIS_W; i--; )
			for (j = VERTRIS_H; j--; )
				if (Vertris[i][j])
				{
					x1 = BOARD_X + SQUARE_SIZE * i;
					y1 = BOARD_Y - SQUARE_SIZE * j;

					draw_vertro(Vertris[i][j], x1, y1 - SQUARE_SIZE, FXVertris[i][j]);

					if (FX)
						do_fx(i, j);
				}
	}

	reset_fx();

	draw_next_vertro();

	textouts();

	acquire_screen();

	output_blit(Buf, screen);
			
	if (Redraw_All)
		Redraw_All = 0;

	release_screen();
	
	Last_Score = Score;
	Last_Vertro = Vertro_Next.type;
}
